from abc import ABC
from functools import reduce
from typing import Iterable, Optional, Callable, Union, Type, Protocol

from exchange_calendars import ExchangeCalendar
from exchange_calendars.exchange_calendar import HolidayCalendar as ExchangeHolidayCalendar
from exchange_calendars.pandas_extensions.holiday import Holiday
import pandas as pd

from exchange_calendars_extensions.holiday import get_monthly_expiry_holiday, DayOfWeekPeriodicHoliday, \
    get_last_day_of_month_holiday
from exchange_calendars_extensions.observance import get_roll_backward_observance


class HolidayCalendar(ExchangeHolidayCalendar):
    def __init__(self, rules):
        super().__init__(rules=rules)

    def holidays(self, start=None, end=None, return_name=False):
        """
        Return all holidays between start and end, inclusive.
        """
        holidays = super().holidays(start=start, end=end, return_name=return_name)
        if return_name:
            return holidays[~holidays.index.duplicated()]
        else:
            return holidays.drop_duplicates()


def get_holiday_calendar_from_timestamps(timestamps: Iterable[pd.Timestamp], name: Optional[str] = None) -> HolidayCalendar:
    """
    Return a calendar with the given holidays.
    """
    rules = [Holiday(name, year=ts.year, month=ts.month, day=ts.day) for ts in list(dict.fromkeys(timestamps))]  # As of Python 3.7, dict preserves insertion order.
    return HolidayCalendar(rules=rules)


def get_holiday_calendar_from_day_of_week(day_of_week: int, name: Optional[str] = None) -> HolidayCalendar:
    """
    Return a calendar with a holiday for each day of the week.
    """
    rules = [DayOfWeekPeriodicHoliday(name, day_of_week)]
    return ExchangeHolidayCalendar(rules=rules)


def merge_calendars(calendars: Iterable[HolidayCalendar]) -> HolidayCalendar:
    """
    Return a calendar with all holidays from the given calendars merged into a single HolidayCalendar.
    """
    x = reduce(lambda x, y: HolidayCalendar(rules=[r for r in x.rules] + [r for r in y.rules]), calendars, HolidayCalendar(rules=[]))
    return x


def get_holidays_calendar(exchange_calendar: ExchangeCalendar) -> HolidayCalendar:
    holiday_calendars = [get_holiday_calendar_from_timestamps(exchange_calendar.adhoc_holidays, name='ad-hoc holiday'),
                         exchange_calendar.regular_holidays]

    # Merge all calendars by reducing the list of calendars into one, calling the merge method on each pair.
    return merge_calendars(holiday_calendars)


def get_special_opens_calendar(exchange_calendar: ExchangeCalendar) -> HolidayCalendar:
    """
    Return a calendar with all special days from the given exchange calendar merged into a single HolidayCalendar.
    """
    holiday_calendars = []

    # Add ad-hoc special opens.
    for item in exchange_calendar.special_opens_adhoc:
        _, definition = item
        holiday_calendars.append(get_holiday_calendar_from_timestamps(definition, name='ad-hoc special open day'))

    # Add regular special open days.
    for item in exchange_calendar.special_opens:
        _, definition = item
        if isinstance(definition, ExchangeHolidayCalendar):
            holiday_calendars.append(definition)
        elif isinstance(definition, int):
            holiday_calendars.append(get_holiday_calendar_from_day_of_week(definition, name='special open day'))

    # Merge all calendars by reducing the list of calendars into one, calling the merge method on each pair.
    return merge_calendars(holiday_calendars)


def get_special_closes_calendar(exchange_calendar: ExchangeCalendar) -> HolidayCalendar:
    """
    Return a calendar with all special days from the given exchange calendar merged into a single HolidayCalendar.
    """
    holiday_calendars = []

    # Add ad-hoc special closes.
    for item in exchange_calendar.special_closes_adhoc:
        _, definition = item
        holiday_calendars.append(get_holiday_calendar_from_timestamps(definition, name='ad-hoc special close day'))

    # Add regular special close days.
    for item in exchange_calendar.special_closes:
        _, definition = item
        if isinstance(definition, ExchangeHolidayCalendar):
            holiday_calendars.append(definition)
        elif isinstance(definition, int):
            holiday_calendars.append(get_holiday_calendar_from_day_of_week(definition, name='special close day'))

    # Merge all calendars by reducing the list of calendars into one, calling the merge method on each pair.
    return merge_calendars(holiday_calendars)


def get_weekend_days_calendar(exchange_calendar: ExchangeCalendar) -> HolidayCalendar:
    rules = [DayOfWeekPeriodicHoliday('weekend day', day_of_week) for day_of_week, v in enumerate(exchange_calendar.weekmask) if v == '0']
    return ExchangeHolidayCalendar(rules=rules)


def get_monthly_expiry_calendar(day_of_week: int, observance: Optional[Callable[[pd.Timestamp], pd.Timestamp]] = None) -> HolidayCalendar:
    """
    Return a calendar with a holiday for each month's expiry, but exclude quarterly expiry days aka quadruple witching.
    """
    rules = [get_monthly_expiry_holiday('monthly expiry', day_of_week, month, observance) for month in [1, 2, 4, 5, 7, 8, 10, 11]]
    return ExchangeHolidayCalendar(rules=rules)


def get_quadruple_witching_calendar(day_of_week: int, observance: Optional[Callable[[pd.Timestamp], pd.Timestamp]] = None) -> HolidayCalendar:
    """
    Return a calendar with a holiday for each quarterly expiry aka quadruple witching.
    """
    rules = [get_monthly_expiry_holiday('quarterly expiry', day_of_week, month, observance) for month in [3, 6, 9, 12]]
    return ExchangeHolidayCalendar(rules=rules)


def get_last_day_of_month_calendar(name: Optional[str] = 'last trading day of month', observance: Optional[Callable[[pd.Timestamp], pd.Timestamp]] = None) -> HolidayCalendar:
    """
    Return a calendar with a holiday for each last day of the month.
    """
    rules = [get_last_day_of_month_holiday(name, i, observance=observance) for i in range(1, 13)]
    return ExchangeHolidayCalendar(rules=rules)


class ExchangeCalendarExtensions(Protocol):

    @property
    def weekend_days(self) -> HolidayCalendar:
        ...  # pragma: no cover

    @property
    def holidays_all(self) -> HolidayCalendar:
        ...  # pragma: no cover

    @property
    def special_opens_all(self) -> HolidayCalendar:
        ...  # pragma: no cover

    @property
    def special_closes_all(self) -> HolidayCalendar:
        ...  # pragma: no cover
    @property
    def monthly_expiries(self) -> Union[HolidayCalendar, None]:
        ...  # pragma: no cover

    @property
    def quarterly_expiries(self) -> Union[HolidayCalendar, None]:
        ...  # pragma: no cover

    @property
    def last_trading_days_of_months(self) -> Union[HolidayCalendar, None]:
        ...  # pragma: no cover

    @property
    def last_regular_trading_days_of_months(self) -> Union[HolidayCalendar, None]:
        ...


class ExtendedExchangeCalendar(ExchangeCalendar, ExchangeCalendarExtensions, ABC):
    ...


def extend_class(cls: Type[ExchangeCalendar], day_of_week_expiry: int = 4) -> type:
    """
    Extend the given ExchangeCalendar class with additional properties.

    This method returns an extended version of the given ExchangeCalendar sub-class with additional properties.
    Specifically, this method adds the following properties:
    - holidays_all: a HolidayCalendar with all holidays, regular and ad-hoc.
    - special_opens_all: a HolidayCalendar with all special open days, regular and ad-hoc.
    - special_closes_all: a HolidayCalendar with all special close days, regular and ad-hoc.
    - weekend_days: a HolidayCalendar with all weekend days.
    - monthly_expiries: a HolidayCalendar with all monthly expiry days.
    - quarterly_expiries: a HolidayCalendar with all quarterly expiry days.
    - last_trading_days_of_months: a HolidayCalendar with the respective last trading day of each month.
    - last_regular_trading_days_of_months: a HolidayCalendar with the respective last regular trading day of each month.

    The properties holidays_all, special_opens_all, and special_closes_all make it more convenient to determine relevant
    days in a given date range since regular and ad-hoc special days are merged into a single calendar.

    The property weekend_days may be useful to determine weekend days in a given date range since parsing the weekmask
    property of the underlying ExchangeCalendar class is avoided.

    The property monthly_expiries returns expiry days for months January, February, April, May, July, August, October,
    and November. For most exchanges, these days are the third Friday or Thursday in the respective month. If that day
    is a holiday or special open/close day, an observance rule determines where the expiry day is shifted to. Typically,
    this will be the previous regular business day. For exchanges that do not observe monthly expiry days, this property
    may throw NotImplementedError.

    Similarly to monthly_expiries, the property quarterly_expiries returns expiry days for months March, June, September,
    and December, also known as quarterly expiries or triple/quadruple witching.

    The property last_trading_days_of_months returns the last trading day of each month. Note that the last trading day
    may be a special open/close day.

    The property last_regular_trading_days_of_months returns the last regular trading day of each month. The only
    difference to last_trading_days_of_months is that this property always returns the last regular trading day and
    never a special open/close day. That is, if the last trading day of a month is a special open/close day, here the
    day is rolled back to the previous regular trading day instead.

    :param cls: the input class to extend.
    :param day_of_week_expiry: the day of the week when expiry days are observed. Defaults to 4, which is Friday.
    :return: the extended class.
    """
    init = cls.__init__

    def __init__(self, *args, **kwargs):
        init(self, *args, **kwargs)
        self._holidays_all = get_holidays_calendar(self)
        self._special_opens_all = get_special_opens_calendar(self)
        self._special_closes_all = get_special_closes_calendar(self)
        special_business_days = merge_calendars([self._special_opens_all, self._special_closes_all])
        self._weekend_days = get_weekend_days_calendar(self)
        weekends_and_holidays = merge_calendars([self._weekend_days, self._holidays_all])
        weekends_holidays_and_special_business_days = merge_calendars([weekends_and_holidays, special_business_days])
        self._monthly_expiry_days = get_monthly_expiry_calendar(day_of_week_expiry, get_roll_backward_observance(weekends_holidays_and_special_business_days))
        self._quarterly_expiry_days = get_quadruple_witching_calendar(day_of_week_expiry, get_roll_backward_observance(weekends_holidays_and_special_business_days))
        self._last_trading_day_of_month = get_last_day_of_month_calendar('last trading day of month', get_roll_backward_observance(weekends_and_holidays))
        self._last_regular_trading_day_of_month = get_last_day_of_month_calendar('last regular trading day of month', get_roll_backward_observance(weekends_holidays_and_special_business_days))

    @property
    def weekend_days(self) -> Union[HolidayCalendar, None]:
        return self._weekend_days

    @property
    def holidays_all(self) -> Union[HolidayCalendar, None]:
        return self._holidays_all

    @property
    def special_opens_all(self) -> Union[HolidayCalendar, None]:
        return self._special_opens_all

    @property
    def special_closes_all(self) -> Union[HolidayCalendar, None]:
        return self._special_closes_all

    @property
    def monthly_expiries(self) -> Union[HolidayCalendar, None]:
        return self._monthly_expiry_days

    @property
    def quarterly_expiries(self) -> Union[HolidayCalendar, None]:
        return self._quarterly_expiry_days

    @property
    def last_trading_days_of_months(self) -> Union[HolidayCalendar, None]:
        return self._last_trading_day_of_month

    @property
    def last_regular_trading_days_of_months(self) -> Union[HolidayCalendar, None]:
        return self._last_regular_trading_day_of_month

    # Use type to create a new class.
    extended = type(cls.__name__ + "Extended", (cls, ExtendedExchangeCalendar), {
        "__init__": __init__,
        "weekend_days": weekend_days,
        "holidays_all": holidays_all,
        "special_opens_all": special_opens_all,
        "special_closes_all": special_closes_all,
        "monthly_expiries": monthly_expiries,
        "quarterly_expiries": quarterly_expiries,
        "last_trading_days_of_months": last_trading_days_of_months,
        "last_regular_trading_days_of_months": last_regular_trading_days_of_months
    })

    return extended



from typing import Callable

import pandas as pd
from exchange_calendars.exchange_calendar import HolidayCalendar


def get_roll_backward_observance(calendar: HolidayCalendar) -> Callable[[pd.Timestamp], pd.Timestamp]:
    """
        Return a function that rolls back a date to the last regular business day, according to a given holiday
        calendar.

        Regular business days are those days that are not defined as holidays by the given holiday calendar. This means,
        for example, that the calendar must also define weekend days as holidays to avoid rolling the date back to a
        such day.
    """

    def f(date: pd.Timestamp) -> pd.Timestamp:
        # Repeat until done.
        while True:
            # A date a few days in the past, relative to the current date.
            limit = date - pd.Timedelta(days=5)

            # Get all holidays between the limit date and the current date.
            holidays = calendar.holidays(limit, date)

            while date >= limit:
                # If the date is a holiday, roll it back one day.
                if date in holidays:
                    date = date - pd.Timedelta(days=1)
                else:
                    # Otherwise, we're done.
                    return date

            # If we get here, we've rolled back more than 5 days. Rinse and repeat.

    return f



from datetime import timedelta
from typing import Optional, Callable

import pandas as pd
from exchange_calendars.pandas_extensions.holiday import Holiday

from exchange_calendars_extensions.offset import LastDayOfMonthOffsetClasses, \
    ThirdDayOfWeekInMonthOffsetClasses


def get_monthly_expiry_holiday(
        name: str,
        day_of_week: int,
        month: int,
        observance: Optional[Callable[[pd.Timestamp], pd.Timestamp]] = None,
        start_date=None,
        end_date=None,
        tz=None) -> Holiday:
    """
    Return a holiday that occurs yearly on the third day of the week in the given month.
    :param name: The name of the holiday.
    :param day_of_week: 0 = Monday, 1 = Tuesday, ..., 6 = Sunday.
    :param month: 1 = January, 2 = February, ..., 12 = December.
    :param observance: A function that takes a datetime and returns a datetime.
    :param start_date: The first date on which this holiday is valid.
    :param end_date: The last date on which this holiday is valid.
    :param tz: The timezone in which to interpret the holiday.
    :return: A Holiday object.
    """
    return Holiday(name, month=1, day=1,
                   offset=ThirdDayOfWeekInMonthOffsetClasses[day_of_week][month](),
                   observance=observance, start_date=start_date, end_date=end_date, tz=tz)


def get_last_day_of_month_holiday(
        name: str,
        month: int,
        observance: Optional[Callable[[pd.Timestamp], pd.Timestamp]] = None,
        start_date=None,
        end_date=None,
        tz=None) -> Holiday:
    """
    Return a holiday that occurs yearly on the last day of the given month.
    :param name: The name of the holiday.
    :param month: 1 = January, 2 = February, ..., 12 = December.
    :param observance: A function that takes a datetime and returns a datetime.
    :param start_date: The first date on which this holiday is valid.
    :param end_date: The last date on which this holiday is valid.
    :param tz: The timezone in which to interpret the holiday.
    :return: A Holiday object.
    """
    return Holiday(name, month=1, day=1,
                   offset=LastDayOfMonthOffsetClasses[month](),
                   observance=observance, start_date=start_date, end_date=end_date, tz=tz)


class DayOfWeekPeriodicHoliday(Holiday):
    """
    A holiday that occurs on a specific day of the week and repeats weekly.
    """

    def __init__(
        self,
        name,
        day_of_week: int,
        start_date=None,
        end_date=None,
        tz=None
    ):
        """
        Constructor.
        :param name: Name of holiday.
        :param day_of_week: 0 = Monday, 1 = Tuesday, ..., 6 = Sunday.
        :param start_date: The first date on which this holiday is valid.
        :param end_date: The last date on which this holiday is valid.
        """
        super().__init__(
            name,
            year=None,
            month=None,
            day=None,
            offset=None,
            observance=None,
            start_date=start_date,
            end_date=end_date,
            days_of_week=None,
            tz=tz
        )
        self.day_of_week = day_of_week

    def _dates(self, start_date, end_date):
        """
        Return a list of dates on which this holiday occurs between start_date and end_date.
        """
        # Determine effective start date.
        if self.start_date is not None:
            start_date = max(start_date, self.start_date.tz_localize(start_date.tz))

        # Determine effective end date.
        if self.end_date is not None:
            end_date = min(end_date, self.end_date.tz_localize(end_date.tz))

        if start_date > end_date:
            return pd.DatetimeIndex([])

        # Get the first date larger or equal to start_date where the day of the week is the same as day_of_week.
        first = start_date + pd.Timedelta(days=(self.day_of_week - start_date.dayofweek) % 7)

        if first > end_date:
            return pd.DatetimeIndex([])

        # Get the last date smaller or equal to end_date where the day of the week is the same as day_of_week.
        last = end_date - pd.Timedelta(days=(end_date.dayofweek - self.day_of_week) % 7)

        # Create a pandas DateTimeIndex with the dates of the holidays.
        dates = pd.date_range(start=first, end=last, freq=timedelta(days=7))

        # Return the dates.
        return dates

    def dates(self, start_date, end_date, return_name=False):
        # Get DateTimeIndex with the dates of the holidays.
        dates = self._dates(start_date, end_date)

        return pd.Series(self.name, index=dates, dtype=pd.DatetimeTZDtype) if return_name else dates



from datetime import datetime, date

import pandas as pd
from pandas._libs.tslibs import localize_pydatetime
from pandas._libs.tslibs.offsets import Easter, apply_wraps

from exchange_calendars_extensions.util import get_month_name, get_day_of_week_name, third_day_of_week_in_month, \
    last_day_in_month


class AbstractHolidayOffset(Easter):

    @staticmethod
    def _is_normalized(dt):
        if dt.hour != 0 or dt.minute != 0 or dt.second != 0 or dt.microsecond != 0:
            # Regardless of whether dt is datetime vs Timestamp
            return False
        if isinstance(dt, pd.Timestamp):
            return dt.nanosecond == 0
        return True

    """
    Auxiliary class for DateOffset instances for the different holidays.
    """

    @property
    def holiday(self):
        """
        Return the Gregorian date for the holiday in a given Gregorian calendar
        year.
        """
        ...  # pragma: no cover

    @apply_wraps
    def _apply(self, other):
        current = self.holiday(other.year)
        current = datetime(current.year, current.month, current.day)
        current = localize_pydatetime(current, other.tzinfo)

        n = self.n
        if n >= 0 and other < current:
            n -= 1
        elif n < 0 and other > current:
            n += 1
        # TODO: Why does this handle the 0 case the opposite of others?

        new = self.holiday(other.year + n)
        new = datetime(
            new.year,
            new.month,
            new.day,
            other.hour,
            other.minute,
            other.second,
            other.microsecond,
        )
        return new

    # backwards compat
    apply = _apply

    def is_on_offset(self, dt):
        if self.normalize and not AbstractHolidayOffset._is_normalized(dt):
            return False
        return date(dt.year, dt.month, dt.day) == self.holiday(dt.year).to_pydate()


def get_third_day_of_week_in_month_offset_class(day_of_week: int, month: int) -> AbstractHolidayOffset:
    """
    Return a new class that represents an offset that when applied to the first day of a year, results in the third given 
    day of the week in the given month.
    
    For example, to get the offset for the third Friday in June, call this function with day_of_week=4 and month=6. On many
    exchanges, this will be the quadruple witching day for the second quarter of the year.
    """

    @property
    def holiday(self):
        """
            Return a function that returns the date of the day for a given year.
        """
        return lambda year: third_day_of_week_in_month(day_of_week, month, year)

    # Get name of day of week.
    day_of_week_name = get_day_of_week_name(day_of_week)

    # Get name of month.
    month_name = get_month_name(month)

    # Create the new class.
    offset = type(f"MonthlyExpiry{month_name}{day_of_week_name}Offset", (AbstractHolidayOffset,), {
        "holiday": holiday,
    })

    # Return the new class.
    return offset


# A dictionary of dictionaries that maps day of week to month to corresponding offset class.
#
# For example, to get the offset class for the third Friday in June, use the following: OffsetClasses[4][6]. To
# instantiate the offset, use the following: OffsetClasses[4][6]().
#
# The offset classes can be used to define typical expiry days (options, futures, et cetera) on exchanges which
# happen on the third Friday or Thursday in a month on most exchanges. The quarterly expiry days in months March, June,
# September, and December are also called quadruple witching.
#
# Currently, includes only Thursdays and Fridays to avoid unnecessarily creating classes that will never be used. Add
# more days here, if required.
ThirdDayOfWeekInMonthOffsetClasses = {day_of_week: {month: get_third_day_of_week_in_month_offset_class(day_of_week, month) for month in range(1, 13)} for day_of_week in range(5)}


def get_last_day_of_month_offset_class(month: int) -> AbstractHolidayOffset:
    @property
    def holiday(self):
        """
            Return a function that returns the last day of the month for a given year.
        """
        return lambda year: last_day_in_month(month, year)

    # Get name of month.
    month_name = get_month_name(month)

    # Create the new class.
    offset = type(f"LastDayOfMonth{month_name}Offset", (AbstractHolidayOffset,), {
        "holiday": holiday,
    })

    # Return the new class.
    return offset


# A dictionary that maps month to corresponding offset class.
LastDayOfMonthOffsetClasses = {month: get_last_day_of_month_offset_class(month) for month in range(1, 13)}



from datetime import timedelta
from functools import reduce
from typing import Iterable, Optional
from datetime import date

from exchange_calendars import ExchangeCalendar
from exchange_calendars.exchange_calendar import HolidayCalendar
from exchange_calendars.pandas_extensions.holiday import Holiday
import pandas as pd


def get_month_name(month: int):
    """
        Convert month to capitalized name of month.

        :param month: Month number (1-12).
        :return: Name of month.
    """
    if month < 1 or month > 12:
        raise ValueError("Month must be between 1 and 12.")

    month_name = \
    ["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November",
     "December"][month - 1]

    return month_name


def get_day_of_week_name(day_of_week: int):
    """
    Convert day of week number to name.

    :param day_of_week: Day of week number (0-6), where 0 is Monday and 6 is Sunday.
    :return: Name of day of week.
    """
    if day_of_week < 0 or day_of_week > 6:
        raise ValueError("Day of week must be between 0 and 6.")

    day_of_week_name = ["Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday"][day_of_week]

    return day_of_week_name


def third_day_of_week_in_month(day_of_week: int, month: int, year: int) -> date:
    """
    Return the third given day of the week in the given month and year.

    :param day_of_week: the day of the week, must be an integer between 0 (Monday) and 6 (Sunday).
    :param year: the year, must be an integer
    :param month: the month of the year, must be an integer between (inclusive) 1 and 12
    :return: the datetime.date representing the third Friday in the given month.
    """
    # The third given day in a month cannot be earlier than the 15th.
    third = date(year, month, 15)

    # Get day of week.
    w = third.weekday()

    # Adjust if necessary.
    if w != day_of_week:
        # Replace just the day of the month, adding a number of days, so that the day of the week is correct.
        third = third.replace(day=(15 + (day_of_week - w) % 7))
    return third


def last_day_in_month(month: int, year: int):
    """
    Return the last day in the given month and year.
    """
    return (date(year, month, 1) + timedelta(days=32)).replace(day=1) - timedelta(days=1)



from exchange_calendars import register_calendar_type
from exchange_calendars.exchange_calendar_asex import ASEXExchangeCalendar
from exchange_calendars.exchange_calendar_xams import XAMSExchangeCalendar
from exchange_calendars.exchange_calendar_xbru import XBRUExchangeCalendar
from exchange_calendars.exchange_calendar_xbud import XBUDExchangeCalendar
from exchange_calendars.exchange_calendar_xcse import XCSEExchangeCalendar
from exchange_calendars.exchange_calendar_xdub import XDUBExchangeCalendar
from exchange_calendars.exchange_calendar_xetr import XETRExchangeCalendar
from exchange_calendars.exchange_calendar_xist import XISTExchangeCalendar
from exchange_calendars.exchange_calendar_xjse import XJSEExchangeCalendar
from exchange_calendars.exchange_calendar_xlis import XLISExchangeCalendar
from exchange_calendars.exchange_calendar_xlon import XLONExchangeCalendar
from exchange_calendars.exchange_calendar_xmil import XMILExchangeCalendar
from exchange_calendars.exchange_calendar_xnys import XNYSExchangeCalendar
from exchange_calendars.exchange_calendar_xosl import XOSLExchangeCalendar
from exchange_calendars.exchange_calendar_xpar import XPARExchangeCalendar
from exchange_calendars.exchange_calendar_xpra import XPRAExchangeCalendar
from exchange_calendars.exchange_calendar_xsto import XSTOExchangeCalendar
from exchange_calendars.exchange_calendar_xswx import XSWXExchangeCalendar
from exchange_calendars.exchange_calendar_xtae import XTAEExchangeCalendar
from exchange_calendars.exchange_calendar_xtse import XTSEExchangeCalendar
from exchange_calendars.exchange_calendar_xwar import XWARExchangeCalendar
from exchange_calendars.exchange_calendar_xwbo import XWBOExchangeCalendar

from .holiday_calendar import extend_class, ExtendedExchangeCalendar


class BMEXExchangeCalendar:
    pass


class XHELEXchangeCalendar:
    pass


class XNASExchangeCalendar:
    pass


_extensions = {
    "ASEX": extend_class(ASEXExchangeCalendar, day_of_week_expiry=4),
    "BMEX": extend_class(BMEXExchangeCalendar, day_of_week_expiry=4),
    "XAMS": extend_class(XAMSExchangeCalendar, day_of_week_expiry=4),
    "XBRU": extend_class(XBRUExchangeCalendar, day_of_week_expiry=4),
    "XBUD": extend_class(XBUDExchangeCalendar, day_of_week_expiry=4),
    "XCSE": extend_class(XCSEExchangeCalendar, day_of_week_expiry=4),
    "XDUB": extend_class(XDUBExchangeCalendar, day_of_week_expiry=4),
    "XETR": extend_class(XETRExchangeCalendar, day_of_week_expiry=4),
    "XHEL": extend_class(XHELEXchangeCalendar, day_of_week_expiry=4),
    "XIST": extend_class(XISTExchangeCalendar, day_of_week_expiry=4),
    "XJSE": extend_class(XJSEExchangeCalendar, day_of_week_expiry=3),
    "XLIS": extend_class(XLISExchangeCalendar, day_of_week_expiry=4),
    "XLON": extend_class(XLONExchangeCalendar, day_of_week_expiry=4),
    "XMIL": extend_class(XMILExchangeCalendar, day_of_week_expiry=4),
    "XNAS": extend_class(XNASExchangeCalendar, day_of_week_expiry=4),
    "XNYS": extend_class(XNYSExchangeCalendar, day_of_week_expiry=4),
    "XOSL": extend_class(XOSLExchangeCalendar, day_of_week_expiry=4),
    "XPAR": extend_class(XPARExchangeCalendar, day_of_week_expiry=4),
    "XPRA": extend_class(XPRAExchangeCalendar, day_of_week_expiry=4),
    "XSTO": extend_class(XSTOExchangeCalendar, day_of_week_expiry=4),
    "XSWX": extend_class(XSWXExchangeCalendar, day_of_week_expiry=4),
    "XTAE": extend_class(XTAEExchangeCalendar, day_of_week_expiry=4),
    "XTSE": extend_class(XTSEExchangeCalendar, day_of_week_expiry=4),
    "XWAR": extend_class(XWARExchangeCalendar, day_of_week_expiry=4),
    "XWBO": extend_class(XWBOExchangeCalendar, day_of_week_expiry=4),
}


def apply_extensions():
    """Apply extensions to exchange_calendars."""
    for k, v in _extensions.items():
        register_calendar_type(k, v, force=True)


def register_extension(name, cls):
    """Set an extension for a given exchange calendar."""
    _extensions[name] = cls


__all__ = ["apply_extensions", "register_extension", "extend_class", "ExtendedExchangeCalendar"]

__version__ = None

try:
    from importlib.metadata import version
    # get version from installed package
    __version__ = version("exchange_calendars_extensions")
except ImportError:
    pass

if __version__ is None:
    try:
        # if package not installed, get version as set when package built
        from ._version import version
    except Exception:
        # If package not installed and not built, leave __version__ as None
        pass
    else:
        __version__ = version

del version
